-------------------------------- MODULE MPGC --------------------------------

(*  A specification that factorizes the common structure 
    of fast and safe modes. The acceptors execute arbitrary proposed
    requests that they read directly from the array pending. The acceptors
    can also not accept any request and set their dState to None. 
    Clients consider their request executed when it appears 
    in the GLB of a RespQuorum of acceptors. The clients abort with 
    an abort value computed from the dstates of an AbortQuorum of 
    acceptors using the operator AbortValues(_).  *)

EXTENDS RDR, Library

CONSTANTS Initial, Acceptor

CONSTANTS RespQuorum, AbortQuorum, AbortValues(_), None
ASSUME None \notin S


VARIABLES status, pending, initVals, dState, accStatus, interface, 
  pastPending
\* abortVals is a history variable
VARIABLE abortVals

INSTANCE SpecLinInterface

vars == <<status, pending, initVals, dState, accStatus, interface, 
  abortVals, pastPending>>
Labels == {"idle", "ready", "pending", "panic", "aborted"}
AcceptorLabels == {"idle","ready","stopped"}

TypeInvariant ==
  /\ \forall p \in P :
    /\ status[p] \in Labels
    /\ pending[p] \in Req
  /\ \forall r \in Acceptor :
    /\ dState[r] \in S \cup {None}
    /\ accStatus[r] \in AcceptorLabels
  /\ initVals \subseteq S
  /\ pastPending \subseteq Req

Init ==
  /\ status = 
    IF Initial
    THEN [p\in P |-> "ready"]
    ELSE  [p\in P |-> "idle"]
  /\ pending = [p \in P |-> Some(Req)]
  /\ initVals = {}
  /\ dState = [r \in Acceptor |-> Bot]
  /\ accStatus = 
    IF Initial
    THEN [r \in Acceptor |-> "ready"]
    ELSE [r \in Acceptor |-> "idle"]
  /\ interface = InterfaceInit
  /\ abortVals = {}
  /\ pastPending = {}

Ini(p) == \E c \in C, v \in S :
  /\ status[p] = "idle"
  /\ pending' = [pending EXCEPT ![p] = <<p,c>>]
  /\ initVals' =  initVals \cup {v}
  /\ status' = [status EXCEPT ![p] = "pending"]
  /\ Initialize(p, c, v)
  /\ pastPending' = pastPending \cup {<<p,c>>}
  /\ UNCHANGED <<dState, accStatus, abortVals>>

Inv(p) == \E c \in C :
  /\ status[p] = "ready"
  /\ pending' = [pending EXCEPT ![p] = <<p,c>>]
  /\ status' = [status EXCEPT ![p] = "pending"]
  /\ Invoke(p, c)
  /\ pastPending' = pastPending \cup {<<p,c>>}
  /\ UNCHANGED <<dState, accStatus, initVals, abortVals>>

AccStates(Q) == 
  {s \in S : \E srv \in Q : s = dState[srv]}

ValidRespQuorum ==
  {Q \in RespQuorum : \A a \in Q : dState[a] # None}
  
Res(p) == 
  /\ status[p] = "pending"
  /\ status' = [status EXCEPT ![p] = "ready"]
  /\ \E Q \in ValidRespQuorum :
    /\ \A srv \in Q : 
      /\ accStatus[srv] # "idle"
    /\ LET glb == GLB(AccStates(Q))
       IN /\  Contains(glb, pending[p])
          /\  Response(p, Output(glb, pending[p]))
  /\ UNCHANGED <<pending, initVals, dState, accStatus, 
    abortVals, pastPending>>

ValidAbortQuorum ==
  {Q \in AbortQuorum : \A a \in Q : dState[a] # None}

Abo(p) ==
  /\ status[p] = "panic"
  /\ 
    \/ \E Q \in ValidAbortQuorum :
      /\ \A srv \in Q : 
        /\ accStatus[srv] = "stopped"
      /\ \E s \in AbortValues([a \in Q |-> dState[a]]) : 
        /\ Abort(p, pending[p][2], s)
        /\ abortVals' = abortVals \cup {s}
    \/ \E Q \in AbortQuorum :
      /\ \A srv \in Q : 
        /\ accStatus[srv] = "stopped"
        /\ dState[srv] = None
      /\ \E s \in initVals : 
        /\ Abort(p, pending[p][2], s)
        /\ abortVals' = abortVals \cup {s}
  /\ status' = [status EXCEPT ![p] = "aborted"]
  /\ UNCHANGED <<pending, initVals, dState, accStatus, pastPending>>

\* We abstract over time: a process can panic at any moment.
Panic(p) ==
  /\ status[p] = "pending"
  /\ status' = [status EXCEPT ![p] = "panic"]
  /\ UNCHANGED <<pending, initVals, dState, accStatus, interface, 
    abortVals, pastPending>>

Pending == 
  {p \in P : status[p] \in {"pending","panic","aborted"}}

\* An Acceptor executes a pending request.
Exec(r) ==
  /\ accStatus[r] = "ready"
  /\ \E req \in pastPending : 
    /\ dState' = [dState EXCEPT ![r] = @ \bullet req]
    /\ dState'[r] \in S \* For TLC
  /\ UNCHANGED <<status, pending, initVals, accStatus, interface, 
    abortVals, pastPending>>

\* An Acceptor sets its local state to one of the init values of the processes.
WakeUp(r) ==
  /\ accStatus[r] = "idle"
  /\ \E iv \in initVals : 
    /\  dState' = [dState EXCEPT ![r] = iv]
    /\  accStatus' = [accStatus EXCEPT ![r] = "ready"]
  /\ UNCHANGED <<status, pending, initVals, interface, abortVals, pastPending>>

Timeout(r) ==
  /\ accStatus[r] = "idle"
  /\ dState' = [dState EXCEPT ![r] = None]
  /\  accStatus' = [accStatus EXCEPT ![r] = "stopped"]
  /\ UNCHANGED <<status, pending, initVals, interface, abortVals, pastPending>>

\* A replica stops when at least one client has panicked
Stop(r) == \E p \in P :
    /\  status[p] \in {"panic","aborted"}
    /\  accStatus[r] = "ready"
    /\  accStatus' = [accStatus EXCEPT ![r] = "stopped"]
    /\  UNCHANGED <<status, pending, initVals, interface, dState, 
      abortVals, pastPending>>

Next ==
    \/  \E p \in P : Ini(p) \/ Inv(p) \/ Res(p) \/ Abo(p) \/ Panic(p)
    \/  \E r \in Acceptor : Exec(r) \/ WakeUp(r) \/ Stop(r) \/ Timeout(r)

Spec == Init /\ [][Next]_vars

=============================================================================
\* Modification History
\* Last modified Fri Jan 24 17:18:16 CET 2014 by losa
\* Created Fri Nov 29 14:12:15 CET 2013 by losa
